# -*- coding: utf-8 -*-
# Generated by Django 1.10.2 on 2016-10-23 23:31
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('cities', '0003_add_verbose_name_and_related_names'),
    ]

    operations = [
        migrations.RenameField(
            model_name='alternativename',
            old_name='language',
            new_name='language_code',
        ),
        migrations.RenameField(
            model_name='country',
            old_name='languages',
            new_name='language_codes',
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import django.contrib.gis.db.models.fields

import swapper


class Migration(migrations.Migration):

    dependencies = [
        swapper.dependency('cities', 'City'),
        swapper.dependency('cities', 'Country'),
    ]

    operations = [
        migrations.CreateModel(
            name='AlternativeName',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=256)),
                ('language', models.CharField(max_length=100)),
                ('is_preferred', models.BooleanField(default=False)),
                ('is_short', models.BooleanField(default=False)),
                ('is_colloquial', models.BooleanField(default=False)),
            ],
        ),
        migrations.CreateModel(
            name='City',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=200, verbose_name='ascii name', db_index=True)),
                ('slug', models.CharField(max_length=200)),
                ('name_std', models.CharField(max_length=200, verbose_name='standard name', db_index=True)),
                ('location', django.contrib.gis.db.models.fields.PointField(srid=4326)),
                ('population', models.IntegerField()),
                ('elevation', models.IntegerField(null=True)),
                ('kind', models.CharField(max_length=10)),
                ('timezone', models.CharField(max_length=40)),
                ('alt_names', models.ManyToManyField(to='cities.AlternativeName')),
            ],
            options={
                'swappable': swapper.swappable_setting('cities', 'City'),
                'verbose_name_plural': 'cities',
            },
        ),
        migrations.CreateModel(
            name='Country',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=200, verbose_name='ascii name', db_index=True)),
                ('slug', models.CharField(max_length=200)),
                ('code', models.CharField(max_length=2, db_index=True)),
                ('code3', models.CharField(max_length=3, db_index=True)),
                ('population', models.IntegerField()),
                ('area', models.IntegerField(null=True)),
                ('currency', models.CharField(max_length=3, null=True)),
                ('currency_name', models.CharField(max_length=50, null=True)),
                ('languages', models.CharField(max_length=250, null=True)),
                ('phone', models.CharField(max_length=20)),
                ('continent', models.CharField(max_length=2)),
                ('tld', models.CharField(max_length=5)),
                ('capital', models.CharField(max_length=100)),
                ('alt_names', models.ManyToManyField(to='cities.AlternativeName')),
                ('neighbours', models.ManyToManyField(related_name='neighbours_rel_+', to=swapper.get_model_name('cities', 'Country'))),
            ],
            options={
                'ordering': ['name'],
                'swappable': swapper.swappable_setting('cities', 'Country'),
                'verbose_name_plural': 'countries',
            },
        ),
        migrations.CreateModel(
            name='District',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=200, verbose_name='ascii name', db_index=True)),
                ('slug', models.CharField(max_length=200)),
                ('name_std', models.CharField(max_length=200, verbose_name='standard name', db_index=True)),
                ('location', django.contrib.gis.db.models.fields.PointField(srid=4326)),
                ('population', models.IntegerField()),
                ('alt_names', models.ManyToManyField(to='cities.AlternativeName')),
                ('city', models.ForeignKey(to=swapper.get_model_name('cities', 'City'))),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='PostalCode',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=200, verbose_name='ascii name', db_index=True)),
                ('slug', models.CharField(max_length=200)),
                ('code', models.CharField(max_length=20)),
                ('location', django.contrib.gis.db.models.fields.PointField(srid=4326)),
                ('region_name', models.CharField(max_length=100, db_index=True)),
                ('subregion_name', models.CharField(max_length=100, db_index=True)),
                ('district_name', models.CharField(max_length=100, db_index=True)),
                ('alt_names', models.ManyToManyField(to='cities.AlternativeName')),
                ('country', models.ForeignKey(related_name='postal_codes', to=swapper.get_model_name('cities', 'Country'))),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Region',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=200, verbose_name='ascii name', db_index=True)),
                ('slug', models.CharField(max_length=200)),
                ('name_std', models.CharField(max_length=200, verbose_name='standard name', db_index=True)),
                ('code', models.CharField(max_length=200, db_index=True)),
                ('alt_names', models.ManyToManyField(to='cities.AlternativeName')),
                ('country', models.ForeignKey(to=swapper.get_model_name('cities', 'Country'))),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Subregion',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=200, verbose_name='ascii name', db_index=True)),
                ('slug', models.CharField(max_length=200)),
                ('name_std', models.CharField(max_length=200, verbose_name='standard name', db_index=True)),
                ('code', models.CharField(max_length=200, db_index=True)),
                ('alt_names', models.ManyToManyField(to='cities.AlternativeName')),
                ('region', models.ForeignKey(to='cities.Region')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='city',
            name='country',
            field=models.ForeignKey(to=swapper.get_model_name('cities', 'Country')),
        ),
        migrations.AddField(
            model_name='city',
            name='region',
            field=models.ForeignKey(blank=True, to='cities.Region', null=True),
        ),
        migrations.AddField(
            model_name='city',
            name='subregion',
            field=models.ForeignKey(blank=True, to='cities.Subregion', null=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.2 on 2016-10-24 12:15
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion

import swapper

from ..conf import ALTERNATIVE_NAME_TYPES


class Migration(migrations.Migration):

    dependencies = [
        ('cities', '0005_add_foreignkeys_to_postalcode'),
        swapper.dependency('cities', 'City'),
    ]

    operations = [
        migrations.AddField(
            model_name='alternativename',
            name='is_historic',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='alternativename',
            name='kind',
            field=models.CharField(choices=ALTERNATIVE_NAME_TYPES, default='name', max_length=4),
        ),
        migrations.AlterField(
            model_name='alternativename',
            name='name',
            field=models.CharField(max_length=255),
        ),
        migrations.AlterField(
            model_name='postalcode',
            name='city',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='postal_codes', to=swapper.get_model_name('cities', 'City')),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.2 on 2016-10-24 18:36
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('cities', '0006_typify_alt_names_and_add_is_historic'),
    ]

    operations = [
        migrations.AddField(
            model_name='country',
            name='currency_symbol',
            field=models.CharField(blank=True, max_length=31, null=True),
        ),
        migrations.AddField(
            model_name='country',
            name='postal_code_format',
            field=models.CharField(default='', max_length=127),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='country',
            name='postal_code_regex',
            field=models.CharField(default='', max_length=255),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='country',
            name='currency_name',
            field=models.CharField(blank=True, max_length=50, null=True),
        ),
    ]



# -*- coding: utf-8 -*-

from importlib import import_module
from collections import defaultdict
from django.conf import settings as django_settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import ugettext_lazy as _

__all__ = [
    'city_types', 'district_types',
    'import_opts', 'import_opts_all', 'HookException', 'settings',
    'ALTERNATIVE_NAME_TYPES', 'CONTINENT_DATA', 'CURRENCY_SYMBOLS',
    'IGNORE_EMPTY_REGIONS', 'INCLUDE_AIRPORT_CODES',
    'INCLUDE_NUMERIC_ALTERNATIVE_NAMES', 'NO_LONGER_EXISTENT_COUNTRY_CODES',
    'SLUGIFY_FUNCTION', 'VALIDATE_POSTAL_CODES',
]

url_bases = {
    'geonames': {
        'dump': 'http://download.geonames.org/export/dump/',
        'zip': 'http://download.geonames.org/export/zip/',
    },
}

files = {
    'country': {
        'filename': 'countryInfo.txt',
        'urls': [url_bases['geonames']['dump'] + '{filename}', ],
        'fields': [
            'code',
            'code3',
            'codeNum',
            'fips',
            'name',
            'capital',
            'area',
            'population',
            'continent',
            'tld',
            'currencyCode',
            'currencyName',
            'phone',
            'postalCodeFormat',
            'postalCodeRegex',
            'languages',
            'geonameid',
            'neighbours',
            'equivalentFips'
        ]
    },
    'region': {
        'filename': 'admin1CodesASCII.txt',
        'urls': [url_bases['geonames']['dump'] + '{filename}', ],
        'fields': [
            'code',
            'name',
            'asciiName',
            'geonameid',
        ]
    },
    'subregion': {
        'filename': 'admin2Codes.txt',
        'urls': [url_bases['geonames']['dump'] + '{filename}', ],
        'fields': [
            'code',
            'name',
            'asciiName',
            'geonameid',
        ]
    },
    'city': {
        'filename': 'cities5000.zip',
        'urls': [url_bases['geonames']['dump'] + '{filename}', ],
        'fields': [
            'geonameid',
            'name',
            'asciiName',
            'alternateNames',
            'latitude',
            'longitude',
            'featureClass',
            'featureCode',
            'countryCode',
            'cc2',
            'admin1Code',
            'admin2Code',
            'admin3Code',
            'admin4Code',
            'population',
            'elevation',
            'gtopo30',
            'timezone',
            'modificationDate'
        ]
    },
    'hierarchy': {
        'filename': 'hierarchy.zip',
        'urls': [url_bases['geonames']['dump'] + '{filename}', ],
        'fields': [
            'parent',
            'child',
            'type',
        ]
    },
    'alt_name': {
        'filename': 'alternateNames.zip',
        'urls': [url_bases['geonames']['dump'] + '{filename}', ],
        'fields': [
            'nameid',
            'geonameid',
            'language',
            'name',
            'isPreferred',
            'isShort',
            'isColloquial',
            'isHistoric',
        ]
    },
    'postal_code': {
        'filename': 'allCountries.zip',
        'urls': [url_bases['geonames']['zip'] + '{filename}', ],
        'fields': [
            'countryCode',
            'postalCode',
            'placeName',
            'admin1Name',
            'admin1Code',
            'admin2Name',
            'admin2Code',
            'admin3Name',
            'admin3Code',
            'latitude',
            'longitude',
            'accuracy',
        ]
    }
}

country_codes = [
    'AD', 'AE', 'AF', 'AG', 'AI', 'AL', 'AM', 'AO', 'AQ', 'AR', 'AS', 'AT', 'AU', 'AW', 'AX', 'AZ',
    'BA', 'BB', 'BD', 'BE', 'BF', 'BG', 'BH', 'BI', 'BJ', 'BL', 'BM', 'BN', 'BO', 'BQ', 'BR', 'BS', 'BT', 'BV', 'BW', 'BY', 'BZ',
    'CA', 'CC', 'CD', 'CF', 'CG', 'CH', 'CI', 'CK', 'CL', 'CM', 'CN', 'CO', 'CR', 'CU', 'CV', 'CW', 'CX', 'CY', 'CZ',
    'DE', 'DJ', 'DK', 'DM', 'DO', 'DZ',
    'EC', 'EE', 'EG', 'EH', 'ER', 'ES', 'ET',
    'FI', 'FJ', 'FK', 'FM', 'FO', 'FR',
    'GA', 'GB', 'GD', 'GE', 'GF', 'GG', 'GH', 'GI', 'GL', 'GM', 'GN', 'GP', 'GQ', 'GR', 'GS', 'GT', 'GU', 'GW', 'GY',
    'HK', 'HM', 'HN', 'HR', 'HT', 'HU',
    'ID', 'IE', 'IL', 'IM', 'IN', 'IO', 'IQ', 'IR', 'IS', 'IT',
    'JE', 'JM', 'JO', 'JP',
    'KE', 'KG', 'KH', 'KI', 'KM', 'KN', 'KP', 'KR', 'XK', 'KW', 'KY', 'KZ',
    'LA', 'LB', 'LC', 'LI', 'LK', 'LR', 'LS', 'LT', 'LU', 'LV', 'LY',
    'MA', 'MC', 'MD', 'ME', 'MF', 'MG', 'MH', 'MK', 'ML', 'MM', 'MN', 'MO', 'MP', 'MQ', 'MR', 'MS', 'MT', 'MU', 'MV', 'MW', 'MX', 'MY', 'MZ',
    'NA', 'NC', 'NE', 'NF', 'NG', 'NI', 'NL', 'NO', 'NP', 'NR', 'NU', 'NZ',
    'OM',
    'PA', 'PE', 'PF', 'PG', 'PH', 'PK', 'PL', 'PM', 'PN', 'PR', 'PS', 'PT', 'PW', 'PY',
    'QA',
    'RE', 'RO', 'RS', 'RU', 'RW',
    'SA', 'SB', 'SC', 'SD', 'SS', 'SE', 'SG', 'SH', 'SI', 'SJ', 'SK', 'SL', 'SM', 'SN', 'SO', 'SR', 'ST', 'SV', 'SX', 'SY', 'SZ',
    'TC', 'TD', 'TF', 'TG', 'TH', 'TJ', 'TK', 'TL', 'TM', 'TN', 'TO', 'TR', 'TT', 'TV', 'TW', 'TZ',
    'UA', 'UG', 'UM', 'US', 'UY', 'UZ',
    'VA', 'VC', 'VE', 'VG', 'VI', 'VN', 'VU',
    'WF', 'WS',
    'YE', 'YT',
    'ZA', 'ZM', 'ZW',
]

_ALTERNATIVE_NAME_TYPES = (
    ('name', _("Name")),
    ('abbr', _("Abbreviation")),
    ('link', _("Link")),
)

_AIRPORT_TYPES = (
    ('iata', _("IATA (Airport) Code")),
    ('icao', _("ICAO (Airport) Code")),
    ('faac', _("FAAC (Airport) Code")),
)

CONTINENT_DATA = {
    'AF': ('Africa', 6255146),
    'AS': ('Asia', 6255147),
    'EU': ('Europe', 6255148),
    'NA': ('North America', 6255149),
    'OC': ('Oceania', 6255151),
    'SA': ('South America', 6255150),
    'AN': ('Antarctica', 6255152),
}

_CURRENCY_SYMBOLS = {
    "AED": "د.إ", "AFN": "؋", "ALL": "L", "AMD": "դր.", "ANG": "ƒ", "AOA": "Kz",
    "ARS": "$", "AUD": "$", "AWG": "ƒ", "AZN": "m",
    "BAM": "KM", "BBD": "$", "BDT": "৳", "BGN": "лв", "BHD": "ب.د", "BIF": "Fr",
    "BMD": "$", "BND": "$", "BOB": "Bs.", "BRL": "R$", "BSD": "$", "BTN": "Nu",
    "BWP": "P", "BYR": "Br", "BZD": "$",
    "CAD": "$", "CDF": "Fr", "CHF": "Fr", "CLP": "$", "CNY": "¥", "COP": "$",
    "CRC": "₡", "CUP": "$", "CVE": "$, Esc", "CZK": "Kč",
    "DJF": "Fr", "DKK": "kr", "DOP": "$", "DZD": "د.ج",
    "EEK": "KR", "EGP": "£,ج.م", "ERN": "Nfk", "ETB": "Br", "EUR": "€",
    "FJD": "$", "FKP": "£",
    "GBP": "£", "GEL": "ლ", "GHS": "₵", "GIP": "£", "GMD": "D", "GNF": "Fr",
    "GTQ": "Q", "GYD": "$",
    "HKD": "$", "HNL": "L", "HRK": "kn", "HTG": "G", "HUF": "Ft",
    "IDR": "Rp", "ILS": "₪", "INR": "₨", "IQD": "ع.د", "IRR": "﷼", "ISK": "kr",
    "JMD": "$", "JOD": "د.ا", "JPY": "¥",
    "KES": "Sh", "KGS": "лв", "KHR": "៛", "KMF": "Fr", "KPW": "₩", "KRW": "₩",
    "KWD": "د.ك", "KYD": "$", "KZT": "Т",
    "LAK": "₭", "LBP": "ل.ل", "LKR": "ரூ", "LRD": "$", "LSL": "L", "LTL": "Lt",
    "LVL": "Ls", "LYD": "ل.د",
    "MAD": "د.م.", "MDL": "L", "MGA": "Ar", "MKD": "ден", "MMK": "K",
    "MNT": "₮", "MOP": "P", "MRO": "UM", "MUR": "₨", "MVR": "ރ.", "MWK": "MK",
    "MXN": "$", "MYR": "RM", "MZN": "MT",
    "NAD": "$", "NGN": "₦", "NIO": "C$", "NOK": "kr", "NPR": "₨", "NZD": "$",
    "OMR": "ر.ع.",
    "PAB": "B/.", "PEN": "S/.", "PGK": "K", "PHP": "₱", "PKR": "₨", "PLN": "zł",
    "PYG": "₲",
    "QAR": "ر.ق",
    "RON": "RON", "RSD": "RSD", "RUB": "р.", "RWF": "Fr",
    "SAR": "ر.س", "SBD": "$", "SCR": "₨", "SDG": "S$", "SEK": "kr", "SGD": "$",
    "SHP": "£", "SLL": "Le", "SOS": "Sh", "SRD": "$", "STD": "Db",
    "SYP": "£, ل.س", "SZL": "L",
    "THB": "฿", "TJS": "ЅМ", "TMT": "m", "TND": "د.ت", "TOP": "T$", "TRY": "₤",
    "TTD": "$", "TWD": "$", "TZS": "Sh",
    "UAH": "₴", "UGX": "Sh", "USD": "$", "UYU": "$", "UZS": "лв",
    "VEF": "Bs", "VND": "₫", "VUV": "Vt",
    "WST": "T",
    "XAF": "Fr", "XCD": "$", "XOF": "Fr", "XPF": "Fr",
    "YER": "﷼",
    "ZAR": "R", "ZMK": "ZK", "ZWL": "$",
}

_NO_LONGER_EXISTENT_COUNTRY_CODES = ['CS', 'AN']

_SLUGIFY_FUNCTION = getattr(django_settings, 'CITIES_SLUGIFY_FUNCTION', 'cities.util.default_slugify')

# See http://www.geonames.org/export/codes.html
city_types = ['PPL', 'PPLA', 'PPLC', 'PPLA2', 'PPLA3', 'PPLA4', 'PPLG']
district_types = ['PPLX']

# Command-line import options
import_opts = [
    'all',
    'country',
    'region',
    'subregion',
    'city',
    'district',
    'alt_name',
    'postal_code',
]

import_opts_all = [
    'country',
    'region',
    'subregion',
    'city',
    'district',
    'alt_name',
    'postal_code',
]


# Raise inside a hook (with an error message) to skip the current line of data.
class HookException(Exception):
    pass


# Hook functions that a plugin class may define
plugin_hooks = [
    'country_pre',     'country_post',  # noqa: E241
    'region_pre',      'region_post',  # noqa: E241
    'subregion_pre',   'subregion_post',  # noqa: E241
    'city_pre',        'city_post',  # noqa: E241
    'district_pre',    'district_post',  # noqa: E241
    'alt_name_pre',    'alt_name_post',  # noqa: E241
    'postal_code_pre', 'postal_code_post',  # noqa: E241
]


def create_settings():
    def get_locales(self):
        if hasattr(django_settings, "CITIES_LOCALES"):
            locales = django_settings.CITIES_LOCALES[:]
        else:
            locales = ['en', 'und']

        try:
            locales.remove('LANGUAGES')
            locales += [e[0] for e in django_settings.LANGUAGES]
        except:
            pass

        return set([e.lower() for e in locales])

    res = type('settings', (), {
        'locales': property(get_locales),
    })

    res.files = files.copy()
    if hasattr(django_settings, "CITIES_FILES"):
        for key in list(django_settings.CITIES_FILES.keys()):
            if 'filenames' in django_settings.CITIES_FILES[key] and 'filename' in django_settings.CITIES_FILES[key]:
                raise ImproperlyConfigured(
                    "Only one key should be specified for '%s': 'filename' of 'filenames'. Both specified instead" % key
                )
            res.files[key].update(django_settings.CITIES_FILES[key])
            if 'filenames' in django_settings.CITIES_FILES[key]:
                del res.files[key]['filename']

    if hasattr(django_settings, "CITIES_DATA_DIR"):
        res.data_dir = django_settings.CITIES_DATA_DIR

    if hasattr(django_settings, "CITIES_POSTAL_CODES"):
        res.postal_codes = set([e.upper() for e in django_settings.CITIES_POSTAL_CODES])
    else:
        res.postal_codes = set(['ALL'])

    return res()


def create_plugins():
    settings.plugins = defaultdict(list)
    for plugin in django_settings.CITIES_PLUGINS:
        module_path, classname = plugin.rsplit('.', 1)
        module = import_module(module_path)
        class_ = getattr(module, classname)
        obj = class_()
        [settings.plugins[hook].append(obj) for hook in plugin_hooks if hasattr(obj, hook)]


settings = create_settings()
if hasattr(django_settings, "CITIES_PLUGINS"):
    create_plugins()

IGNORE_EMPTY_REGIONS = getattr(django_settings, 'CITIES_IGNORE_EMPTY_REGIONS', False)

# Users may way to import historical countries
NO_LONGER_EXISTENT_COUNTRY_CODES = getattr(
    django_settings, 'CITIES_NO_LONGER_EXISTENT_COUNTRY_CODES',
    _NO_LONGER_EXISTENT_COUNTRY_CODES)

# Users may not want to include airport codes as alternative city names
INCLUDE_AIRPORT_CODES = getattr(django_settings, 'CITIES_INCLUDE_AIRPORT_CODES', True)

if INCLUDE_AIRPORT_CODES:
    _ALTERNATIVE_NAME_TYPES += _AIRPORT_TYPES

# A `Choices` object (from `django-model-utils`)
ALTERNATIVE_NAME_TYPES = getattr(django_settings, 'CITIES_ALTERNATIVE_NAME_TYPES', _ALTERNATIVE_NAME_TYPES)

INCLUDE_NUMERIC_ALTERNATIVE_NAMES = getattr(django_settings, 'CITIES_INCLUDE_NUMERIC_ALTERNATIVE_NAMES', True)

# Allow users to override specified contents
CONTINENT_DATA.update(getattr(django_settings, 'CITIES_CONTINENT_DATA', {}))

CURRENCY_SYMBOLS = getattr(django_settings, 'CITIES_CURRENCY_SYMBOLS', _CURRENCY_SYMBOLS)

module_name, _, function_name = _SLUGIFY_FUNCTION.rpartition('.')
SLUGIFY_FUNCTION = getattr(import_module(module_name), function_name)

# Users who want better postal codes can flip this on (developers of
# django-cities itself probably will), but most probably won't want to
VALIDATE_POSTAL_CODES = getattr(django_settings, 'CITIES_VALIDATE_POSTAL_CODES', False)



code_map = {
    'AB': '01',
    'BC': '02',
    'MB': '03',
    'NB': '04',
    'NL': '05',
    'NS': '07',
    'ON': '08',
    'PE': '09',
    'QC': '10',
    'SK': '11',
    'YT': '12',
    'NT': '13',
    'NU': '14',
}


class Plugin:
    def postal_code_pre(self, parser, item):
        country_code = item['countryCode']
        if country_code != 'CA':
            return
        item['admin1Code'] = code_map[item['admin1Code']]



"""
GeoNames city data import script.
Requires the following files:

http://download.geonames.org/export/dump/
- Countries:            countryInfo.txt
- Regions:              admin1CodesASCII.txt
- Subregions:           admin2Codes.txt
- Cities:               cities5000.zip
- Districts:            hierarchy.zip
- Localization:         alternateNames.zip

http://download.geonames.org/export/zip/
- Postal Codes:         allCountries.zip
"""

from __future__ import print_function

import io
import json
import logging
import math
import os
import re
import sys
import zipfile

try:
    from urllib.request import urlopen
except ImportError:
    from urllib import urlopen

from itertools import chain
from optparse import make_option
from swapper import load_model
from tqdm import tqdm

from django import VERSION as django_version
from django.contrib.gis.gdal.envelope import Envelope
from django.contrib.gis.geos import Point
try:
    from django.contrib.gis.db.models.functions import Distance
except ImportError:
    pass
from django.core.management.base import BaseCommand
from django.db import transaction
from django.db.models import Q
from django.db.models import CharField, ForeignKey

from ...conf import (city_types, district_types, import_opts, import_opts_all,
                     HookException, settings, CURRENCY_SYMBOLS,
                     IGNORE_EMPTY_REGIONS, INCLUDE_AIRPORT_CODES,
                     INCLUDE_NUMERIC_ALTERNATIVE_NAMES,
                     NO_LONGER_EXISTENT_COUNTRY_CODES, VALIDATE_POSTAL_CODES)
from ...models import (Region, Subregion, District, PostalCode, AlternativeName)
from ...util import geo_distance


# Interpret all files as utf-8
if sys.version_info < (3,):
    reload(sys)  # noqa: F821
    sys.setdefaultencoding('utf-8')

# Load swappable models
Continent = load_model('cities', 'Continent')
Country = load_model('cities', 'Country')
City = load_model('cities', 'City')


# Only log errors during Travis tests
LOGGER_NAME = os.environ.get('TRAVIS_LOGGER_NAME', 'cities')

# TODO: Remove backwards compatibility once django-cities requires Django 1.7
# or 1.8 LTS.
# _transact = (transaction.commit_on_success if django_version < (1, 6) else
#              transaction.atomic)


class Command(BaseCommand):
    if hasattr(settings, 'data_dir'):
        data_dir = settings.data_dir
    else:
        app_dir = os.path.normpath(os.path.dirname(os.path.realpath(__file__)) + '/../..')
        data_dir = os.path.join(app_dir, 'data')
    logger = logging.getLogger(LOGGER_NAME)

    if django_version < (1, 8):
        option_list = getattr(BaseCommand, 'option_list', ()) + (
            make_option(
                '--force',
                action='store_true',
                default=False,
                help='Import even if files are up-to-date.'),
            make_option(
                '--import',
                metavar="DATA_TYPES",
                default='all',
                help='Selectively import data. Comma separated list of data ' +
                     'types: ' + str(import_opts).replace("'", '')),
            make_option(
                '--flush',
                metavar="DATA_TYPES",
                default='',
                help="Selectively flush data. Comma separated list of data types."),
        )

    def add_arguments(self, parser):
        parser.add_argument(
            '--force',
            action='store_true',
            default=False,
            dest="force",
            help='Import even if files are up-to-date.'
        )
        parser.add_argument(
            '--import',
            metavar="DATA_TYPES",
            default='all',
            dest="import",
            help='Selectively import data. Comma separated list of data ' +
                 'types: ' + str(import_opts).replace("'", '')
        )
        parser.add_argument(
            '--flush',
            metavar="DATA_TYPES",
            default='',
            dest="flush",
            help="Selectively flush data. Comma separated list of data types."
        )

    @transaction.atomic
    def handle(self, *args, **options):
        self.download_cache = {}
        self.options = options

        self.force = self.options['force']

        self.flushes = [e for e in self.options.get('flush', '').split(',') if e]
        if 'all' in self.flushes:
            self.flushes = import_opts_all
        for flush in self.flushes:
            func = getattr(self, "flush_" + flush)
            func()

        self.imports = [e for e in self.options.get('import', '').split(',') if e]
        if 'all' in self.imports:
            self.imports = import_opts_all
        if self.flushes:
            self.imports = []
        for import_ in self.imports:
            func = getattr(self, "import_" + import_)
            func()

    def call_hook(self, hook, *args, **kwargs):
        if hasattr(settings, 'plugins'):
            for plugin in settings.plugins[hook]:
                try:
                    func = getattr(plugin, hook)
                    func(self, *args, **kwargs)
                except HookException as e:
                    error = str(e)
                    if error:
                        self.logger.error(error)
                    return False
        return True

    def download(self, filekey, key_index=None):
        if key_index is None:
            filename = settings.files[filekey]['filename']
        else:
            filename = settings.files[filekey]['filenames'][key_index]

        web_file = None
        urls = [e.format(filename=filename) for e in settings.files[filekey]['urls']]
        for url in urls:
            try:
                web_file = urlopen(url)
                if 'html' in web_file.headers['Content-Type']:
                    raise Exception()
                self.logger.debug("Downloaded: {}".format(url))
                break
            except:
                web_file = None
                continue
        else:
            self.logger.error("Web file not found: %s. Tried URLs:\n%s", filename, '\n'.join(urls))

        if web_file is not None:
            self.logger.debug("Saving: {}/{}".format(self.data_dir, filename))
            if not os.path.exists(self.data_dir):
                os.makedirs(self.data_dir)
            file = io.open(os.path.join(self.data_dir, filename), 'wb')
            file.write(web_file.read())
            file.close()
        elif not os.path.exists(os.path.join(self.data_dir, filename)):
            raise Exception("File not found and download failed: {} [{}]".format(filename, url))

    def get_data(self, filekey):
        if 'filename' in settings.files[filekey]:
            filenames = [settings.files[filekey]['filename']]
        else:
            filenames = settings.files[filekey]['filenames']

        for filename in filenames:
            name, ext = filename.rsplit('.', 1)
            if (ext == 'zip'):
                filepath = os.path.join(self.data_dir, filename)
                zip_member = zipfile.ZipFile(filepath).open(name + '.txt', 'r')
                file_obj = io.TextIOWrapper(zip_member, encoding='utf-8')
            else:
                file_obj = io.open(os.path.join(self.data_dir, filename),
                                   'r', encoding='utf-8')

            for row in file_obj:
                if not row.startswith('#'):
                    yield dict(list(zip(settings.files[filekey]['fields'],
                                        row.rstrip('\n').split("\t"))))

    def parse(self, data):
        for line in data:
            if len(line) < 1 or line[0] == '#':
                continue
            items = [e.strip() for e in line.split('\t')]
            yield items

    def import_country(self):
        self.download('country')
        data = self.get_data('country')

        total = sum(1 for _ in data) - len(NO_LONGER_EXISTENT_COUNTRY_CODES)

        data = self.get_data('country')

        neighbours = {}
        countries = {}

        continents = {c.code: c for c in Continent.objects.all()}

        # If the continent attribute on Country is a ForeignKey, import
        # continents as ForeignKeys to the Continent models, otherwise assume
        # they are still the CharField(max_length=2) and import them the old way
        import_continents_as_fks = type(Country._meta.get_field('continent')) == ForeignKey

        for item in tqdm([d for d in data if d['code'] not in NO_LONGER_EXISTENT_COUNTRY_CODES],
                         total=total,
                         desc="Importing countries"):
            if not self.call_hook('country_pre', item):
                continue

            try:
                country_id = int(item['geonameid'])
            except:
                self.logger.warning("Country has no geonameid: {} -- skipping".format(item))
                continue

            defaults = {
                'name': item['name'],
                'code': item['code'],
                'code3': item['code3'],
                'population': item['population'],
                'continent': continents[item['continent']] if import_continents_as_fks else item['continent'],
                'tld': item['tld'][1:],  # strip the leading .
                'phone': item['phone'],
                'currency': item['currencyCode'],
                'currency_name': item['currencyName'],
                'capital': item['capital'],
                'area': int(float(item['area'])) if item['area'] else None,
            }

            if hasattr(Country, 'language_codes'):
                defaults['language_codes'] = item['languages']
            elif hasattr(Country, 'languages') and type(getattr(Country, 'languages')) == CharField:
                defaults['languages'] = item['languages']

            # These fields shouldn't impact saving older models (that don't
            # have these attributes)
            try:
                defaults['currency_symbol'] = CURRENCY_SYMBOLS.get(item['currencyCode'], None)
                defaults['postal_code_format'] = item['postalCodeFormat']
                defaults['postal_code_regex'] = item['postalCodeRegex']
            except AttributeError:
                pass

            # Make importing countries idempotent
            country, created = Country.objects.update_or_create(id=country_id, defaults=defaults)

            self.logger.debug("%s country '%s'",
                              "Added" if created else "Updated",
                              defaults['name'])

            neighbours[country] = item['neighbours'].split(",")
            countries[country.code] = country

            if not self.call_hook('country_post', country, item):
                continue

        for country, neighbour_codes in tqdm(list(neighbours.items()),
                                             total=len(neighbours),
                                             desc="Importing country neighbours"):
            neighbours = [x for x in [countries.get(x) for x in neighbour_codes if x] if x]
            country.neighbours.add(*neighbours)

    def build_country_index(self):
        if hasattr(self, 'country_index'):
            return

        self.country_index = {}
        for obj in tqdm(Country.objects.all(),
                        total=Country.objects.all().count(),
                        desc="Building country index"):
            self.country_index[obj.code] = obj

    def import_region(self):
        self.download('region')
        data = self.get_data('region')

        self.build_country_index()

        total = sum(1 for _ in data)

        data = self.get_data('region')

        countries_not_found = {}
        for item in tqdm(data, total=total, desc="Importing regions"):
            if not self.call_hook('region_pre', item):
                continue

            try:
                region_id = int(item['geonameid'])
            except:
                self.logger.warning("Region has no geonameid: {} -- skipping".format(item))
                continue

            country_code, region_code = item['code'].split(".")

            defaults = {
                'name': item['name'],
                'name_std': item['asciiName'],
                'code': region_code,
            }

            try:
                defaults['country'] = self.country_index[country_code]
            except KeyError:
                countries_not_found.setdefault(country_code, []).append(defaults['name'])
                self.logger.warning("Region: %s: Cannot find country: %s -- skipping",
                                    defaults['name'], country_code)
                continue

            region, created = Region.objects.update_or_create(id=region_id, defaults=defaults)

            if not self.call_hook('region_post', region, item):
                continue

            self.logger.debug("%s region: %s, %s",
                              "Added" if created else "Updated",
                              item['code'], region)

        if countries_not_found:
            countries_not_found_file = os.path.join(self.data_dir, 'countries_not_found.json')
            try:
                with open(countries_not_found_file, 'w+') as fp:
                    json.dump(countries_not_found, fp)
            except Exception as e:
                self.logger.warning("Unable to write log file '{}': {}".format(
                                    countries_not_found_file, e))

    def build_region_index(self):
        if hasattr(self, 'region_index'):
            return

        self.region_index = {}
        for obj in tqdm(chain(Region.objects.all().prefetch_related('country'),
                              Subregion.objects.all().prefetch_related('region__country')),
                        total=Region.objects.all().count() + Subregion.objects.all().count(),
                        desc="Building region index"):
            self.region_index[obj.full_code()] = obj

    def import_subregion(self):
        self.download('subregion')
        data = self.get_data('subregion')

        total = sum(1 for _ in data)

        data = self.get_data('subregion')

        self.build_country_index()
        self.build_region_index()

        regions_not_found = {}
        for item in tqdm(data, total=total, desc="Importing subregions"):
            if not self.call_hook('subregion_pre', item):
                continue

            try:
                subregion_id = int(item['geonameid'])
            except:
                self.logger.warning("Subregion has no geonameid: {} -- skipping".format(item))
                continue

            country_code, region_code, subregion_code = item['code'].split(".")

            defaults = {
                'name': item['name'],
                'name_std': item['asciiName'],
                'code': subregion_code,
            }

            try:
                defaults['region'] = self.region_index[country_code + "." + region_code]
            except:
                regions_not_found.setdefault(country_code, {})
                regions_not_found[country_code].setdefault(region_code, []).append(defaults['name'])
                self.logger.debug("Subregion: %s %s: Cannot find region",
                                  item['code'], defaults['name'])
                continue

            subregion, created = Subregion.objects.update_or_create(id=subregion_id, defaults=defaults)

            if not self.call_hook('subregion_post', subregion, item):
                continue

            self.logger.debug("%s subregion: %s, %s",
                              "Added" if created else "Updated",
                              item['code'], subregion)

        if regions_not_found:
            regions_not_found_file = os.path.join(self.data_dir, 'regions_not_found.json')
            try:
                with open(regions_not_found_file, 'w+') as fp:
                    json.dump(regions_not_found, fp)
            except Exception as e:
                self.logger.warning("Unable to write log file '{}': {}".format(
                                    regions_not_found_file, e))

        del self.region_index

    def import_city(self):
        self.download('city')
        data = self.get_data('city')

        total = sum(1 for _ in data)

        data = self.get_data('city')

        self.build_country_index()
        self.build_region_index()

        for item in tqdm(data, total=total, desc="Importing cities"):
            if not self.call_hook('city_pre', item):
                continue

            if item['featureCode'] not in city_types:
                continue

            try:
                city_id = int(item['geonameid'])
            except:
                self.logger.warning("City has no geonameid: {} -- skipping".format(item))
                continue

            defaults = {
                'name': item['name'],
                'kind': item['featureCode'],
                'name_std': item['asciiName'],
                'location': Point(float(item['longitude']), float(item['latitude'])),
                'population': int(item['population']),
                'timezone': item['timezone'],
            }

            try:
                defaults['elevation'] = int(item['elevation'])
            except:
                pass

            country_code = item['countryCode']
            try:
                country = self.country_index[country_code]
                defaults['country'] = country
            except:
                self.logger.warning("City: %s: Cannot find country: %s -- skipping",
                                    item['name'], country_code)
                continue

            region_code = item['admin1Code']
            try:
                region = self.region_index[country_code + "." + region_code]
                defaults['region'] = region
            except:
                if IGNORE_EMPTY_REGIONS:
                    defaults['region'] = None
                else:
                    self.logger.debug("%s: %s: Cannot find region: %s -- skipping",
                                      country_code, item['name'], region_code)
                    continue

            subregion_code = item['admin2Code']
            try:
                subregion = self.region_index[country_code + "." + region_code + "." + subregion_code]
                defaults['subregion'] = subregion
            except KeyError:
                try:
                    with transaction.atomic():
                        defaults['subregion'] = Subregion.objects.get(
                            Q(name=subregion_code) |
                            Q(name=subregion_code.replace(' (undefined)', '')),
                            region=defaults['region'])
                except Subregion.DoesNotExist:
                    try:
                        with transaction.atomic():
                            defaults['subregion'] = Subregion.objects.get(
                                Q(name_std=subregion_code) |
                                Q(name_std=subregion_code.replace(' (undefined)', '')),
                                region=defaults['region'])
                    except Subregion.DoesNotExist:
                        if subregion_code:
                            self.logger.debug("%s: %s: Cannot find subregion: %s",
                                              country_code, item['name'], subregion_code)
                        defaults['subregion'] = None

            city, created = City.objects.update_or_create(id=city_id, defaults=defaults)

            if not self.call_hook('city_post', city, item):
                continue

            self.logger.debug("%s city: %s",
                              "Added" if created else "Updated", city)

    def build_hierarchy(self):
        if hasattr(self, 'hierarchy') and self.hierarchy:
            return

        self.download('hierarchy')
        data = self.get_data('hierarchy')

        total = sum(1 for _ in data)

        data = self.get_data('hierarchy')

        self.hierarchy = {}
        for item in tqdm(data, total=total, desc="Building hierarchy index"):
            parent_id = int(item['parent'])
            child_id = int(item['child'])
            self.hierarchy[child_id] = parent_id

    def import_district(self):
        self.download('city')
        data = self.get_data('city')

        total = sum(1 for _ in data)

        data = self.get_data('city')

        self.build_country_index()
        self.build_region_index()
        self.build_hierarchy()

        city_index = {}
        for obj in tqdm(City.objects.all(), total=City.objects.all().count(),
                        desc="Building city index"):
            city_index[obj.id] = obj

        for item in tqdm(data, total=total, desc="Importing districts"):
            if not self.call_hook('district_pre', item):
                continue

            _type = item['featureCode']
            if _type not in district_types:
                continue

            defaults = {
                'name': item['name'],
                'name_std': item['asciiName'],
                'location': Point(float(item['longitude']), float(item['latitude'])),
                'population': int(item['population']),
            }

            if hasattr(District, 'code'):
                defaults['code'] = item['admin3Code'],

            # Find city
            city = None
            try:
                city = city_index[self.hierarchy[item['geonameid']]]
            except:
                self.logger.debug("District: %d %s: Cannot find city in hierarchy, using nearest", item['geonameid'], defaults['name'])
                city_pop_min = 100000
                # we are going to try to find closet city using native
                # database .distance(...) query but if that fails then
                # we fall back to degree search, MYSQL has no support
                # and Spatialite with SRID 4236.
                try:
                    if django_version < (1, 9):
                        city = City.objects.filter(population__gt=city_pop_min)\
                                   .distance(defaults['location'])\
                                   .order_by('distance')[0]
                    else:
                        city = City.objects.filter(population__gt=city_pop_min)\
                            .annotate(distance=Distance('location', defaults['location']))\
                            .order_by('distance')[0]
                except:  # TODO: Restrict what this catches
                    self.logger.warning(
                        "District: %s: DB backend does not support native '.distance(...)' query "
                        "falling back to two degree search",
                        defaults['name']
                    )
                    search_deg = 2
                    min_dist = float('inf')
                    bounds = Envelope(
                        defaults['location'].x - search_deg, defaults['location'].y - search_deg,
                        defaults['location'].x + search_deg, defaults['location'].y + search_deg)
                    for e in City.objects.filter(population__gt=city_pop_min).filter(
                            location__intersects=bounds.wkt):
                        dist = geo_distance(defaults['location'], e.location)
                        if dist < min_dist:
                            min_dist = dist
                            city = e

            if not city:
                self.logger.warning("District: %s: Cannot find city -- skipping", defaults['name'])
                continue

            defaults['city'] = city

            try:
                with transaction.atomic():
                    district = District.objects.get(city=defaults['city'], name=defaults['name'])
            except District.DoesNotExist:
                # If the district doesn't exist, create it with the geonameid
                # as its id
                district, created = District.objects.update_or_create(id=item['geonameid'], defaults=defaults)
            else:
                # Since the district already exists, but doesn't have its
                # geonameid as its id, we need to update all of its attributes
                # *except* for its id
                for key, value in defaults.items():
                    setattr(district, key, value)
                district.save()
                created = False

            if not self.call_hook('district_post', district, item):
                continue

            self.logger.debug("%s district: %s", "Added" if created else "Updated", district)

    def import_alt_name(self):
        self.download('alt_name')
        data = self.get_data('alt_name')

        total = sum(1 for _ in data)

        data = self.get_data('alt_name')

        geo_index = {}
        for type_ in (Country, Region, Subregion, City, District):
            plural_type_name = '{}s'.format(type_.__name__) if type_.__name__[-1] != 'y' else '{}ies'.format(type_.__name__[:-1])
            for obj in tqdm(type_.objects.all(),
                            total=type_.objects.all().count(),
                            desc="Building geo index for {}".format(plural_type_name.lower())):
                geo_index[obj.id] = {
                    'type': type_,
                    'object': obj,
                }

        for item in tqdm(data, total=total, desc="Importing data for alternative names"):
            if not self.call_hook('alt_name_pre', item):
                continue

            # Only get names for languages in use
            locale = item['language']
            if not locale:
                locale = 'und'
            if locale not in settings.locales and 'all' not in settings.locales:
                self.logger.debug(
                    "Alternative name with language [{}]: {} "
                    "({}) -- skipping".format(
                        item['language'], item['name'], item['nameid']))
                continue

            # Check if known geo id
            geo_id = int(item['geonameid'])
            try:
                geo_info = geo_index[geo_id]
            except:
                continue

            try:
                alt_id = int(item['nameid'])
            except KeyError:
                self.logger.warning("Alternative name has no nameid: {} -- skipping".format(item))
                continue

            try:
                alt = AlternativeName.objects.get(id=alt_id)
            except AlternativeName.DoesNotExist:
                alt = AlternativeName(id=alt_id)

            alt.name = item['name']
            alt.is_preferred = bool(item['isPreferred'])
            alt.is_short = bool(item['isShort'])
            try:
                alt.language_code = locale
            except:
                alt.language = locale

            try:
                int(item['name'])
            except:
                pass
            else:
                if not INCLUDE_NUMERIC_ALTERNATIVE_NAMES:
                    self.logger.debug(
                        "Trying to add a numeric alternative name to {} ({}): {} -- skipping".format(
                            geo_info['object'].name,
                            geo_info['type'].__name__,
                            item['name']))
                    continue
            alt.is_historic = True if ((item['isHistoric']and
                                        item['isHistoric'] != '\n') or
                                       locale == 'fr_1793') else False

            if locale == 'post':
                try:
                    if geo_index[item['geonameid']]['type'] == Region:
                        region = geo_index[item['geonameid']]['object']
                        PostalCode.objects.get_or_create(
                            code=item['name'],
                            country=region.country,
                            region=region,
                            region_name=region.name)
                    elif geo_index[item['geonameid']]['type'] == Subregion:
                        subregion = geo_index[item['geonameid']]['object']
                        PostalCode.objects.get_or_create(
                            code=item['name'],
                            country=subregion.region.country,
                            region=subregion.region,
                            subregion=subregion,
                            region_name=subregion.region.name,
                            subregion_name=subregion.name)
                    elif geo_index[item['geonameid']]['type'] == City:
                        city = geo_index[item['geonameid']]['object']
                        PostalCode.objects.get_or_create(
                            code=item['name'],
                            country=city.country,
                            region=city.region,
                            subregion=city.subregion,
                            region_name=city.region.name,
                            subregion_name=city.subregion.name)
                except KeyError:
                    pass

                continue

            if hasattr(alt, 'kind'):
                if (locale in ('abbr', 'link', 'name') or
                   INCLUDE_AIRPORT_CODES and locale in ('iana', 'icao', 'faac')):
                    alt.kind = locale
                elif locale not in settings.locales and 'all' not in settings.locales:
                    self.logger.debug("Unknown alternative name type: {} -- skipping".format(locale))
                    continue

            alt.save()
            geo_info['object'].alt_names.add(alt)

            if not self.call_hook('alt_name_post', alt, item):
                continue

            self.logger.debug("Added alt name: %s, %s", locale, alt)

    def build_postal_code_regex_index(self):
        if hasattr(self, 'postal_code_regex_index') and self.postal_code_regex_index:
            return

        self.build_country_index()

        self.postal_code_regex_index = {}
        for code, country in tqdm(self.country_index.items(),
                                  total=len(self.country_index),
                                  desc="Building postal code regex index"):
            try:
                self.postal_code_regex_index[code] = re.compile(country.postal_code_regex)
            except Exception as e:
                self.logger.error("Couldn't compile postal code regex for {}: {}".format(country.code, e.args))
                self.postal_code_regex_index[code] = ''

    def import_postal_code(self):
        self.download('postal_code')
        data = self.get_data('postal_code')

        total = sum(1 for _ in data)

        data = self.get_data('postal_code')

        self.build_country_index()
        self.build_region_index()
        if VALIDATE_POSTAL_CODES:
            self.build_postal_code_regex_index()

        districts_to_delete = []

        query_statistics = [0 for i in range(8)]
        num_existing_postal_codes = PostalCode.objects.count()
        if num_existing_postal_codes == 0:
            self.logger.debug("Zero postal codes found - using only-create "
                              "postal code optimization")
        for item in tqdm(data, total=total, desc="Importing postal codes"):
            if not self.call_hook('postal_code_pre', item):
                continue

            country_code = item['countryCode']
            if country_code not in settings.postal_codes and 'ALL' not in settings.postal_codes:
                continue

            try:
                code = item['postalCode']
            except KeyError:
                self.logger.warning("Postal code has no code: {} -- skipping".format(item))
                continue

            # Find country
            try:
                country = self.country_index[country_code]
            except:
                self.logger.warning("Postal code '%s': Cannot find country: %s -- skipping", code, country_code)
                continue

            # Validate postal code against the country
            code = item['postalCode']
            if VALIDATE_POSTAL_CODES and self.postal_code_regex_index[country_code].match(code) is None:
                self.logger.warning("Postal code didn't validate: {} ({})".format(code, country_code))
                continue

            reg_name_q = Q(region_name__iexact=item['admin1Name'])
            subreg_name_q = Q(subregion_name__iexact=item['admin2Name'])
            dst_name_q = Q(district_name__iexact=item['admin3Name'])

            if hasattr(PostalCode, 'region'):
                reg_name_q |= Q(region__code=item['admin1Code'])

            if hasattr(PostalCode, 'subregion'):
                subreg_name_q |= Q(subregion__code=item['admin2Code'])

            if hasattr(PostalCode, 'district') and hasattr(District, 'code'):
                dst_name_q |= Q(district__code=item['admin3Code'])

            try:
                location = Point(float(item['longitude']),
                                 float(item['latitude']))
            except:
                location = None

            if len(item['placeName']) >= 200:
                self.logger.warning("Postal code name has more than 200 characters: {}".format(item))

            if num_existing_postal_codes > 0:
                postal_code_args = (
                    {
                        'args': (reg_name_q, subreg_name_q, dst_name_q),
                        'country': country,
                        'code': code,
                        'location': location,
                    }, {
                        'args': (reg_name_q, subreg_name_q, dst_name_q),
                        'country': country,
                        'code': code,
                    }, {
                        'args': (reg_name_q, subreg_name_q, dst_name_q),
                        'country': country,
                        'code': code,
                        'name__iexact': re.sub("'", '', item['placeName']),
                    }, {
                        'args': tuple(),
                        'country': country,
                        'region__code': item['admin1Code'],
                    }, {
                        'args': tuple(),
                        'country': country,
                        'code': code,
                        'name': item['placeName'],
                        'region__code': item['admin1Code'],
                        'subregion__code': item['admin2Code'],
                    }, {
                        'args': tuple(),
                        'country': country,
                        'code': code,
                        'name': item['placeName'],
                        'region__code': item['admin1Code'],
                        'subregion__code': item['admin2Code'],
                        'district__code': item['admin3Code'],
                    }, {
                        'args': tuple(),
                        'country': country,
                        'code': code,
                        'name': item['placeName'],
                        'region_name': item['admin1Name'],
                        'subregion_name': item['admin2Name'],
                    }, {
                        'args': tuple(),
                        'country': country,
                        'code': code,
                        'name': item['placeName'],
                        'region_name': item['admin1Name'],
                        'subregion_name': item['admin2Name'],
                        'district_name': item['admin3Name'],
                    }
                )

                # We do this so we don't have to deal with exceptions being thrown
                # in the middle of transactions
                for args_dict in postal_code_args:
                    num_pcs = PostalCode.objects.filter(
                        *args_dict['args'],
                        **{k: v for k, v in args_dict.items() if k != 'args'})\
                        .count()
                    if num_pcs == 1:
                        pc = PostalCode.objects.get(
                            *args_dict['args'],
                            **{k: v for k, v in args_dict.items() if k != 'args'})
                        break
                    elif num_pcs > 1:
                        pcs = PostalCode.objects.filter(
                            *args_dict['args'],
                            **{k: v for k, v in args_dict.items() if k != 'args'})
                        self.logger.debug("item: {}\nresults: {}".format(item, pcs))
                        # Raise a MultipleObjectsReturned exception
                        PostalCode.objects.get(
                            *args_dict['args'],
                            **{k: v for k, v in args_dict.items() if k != 'args'})
                else:
                    self.logger.debug("Creating postal code: {}".format(item))
                    pc = PostalCode(
                        country=country,
                        code=code,
                        name=item['placeName'],
                        region_name=item['admin1Name'],
                        subregion_name=item['admin2Name'],
                        district_name=item['admin3Name'])
            else:
                self.logger.debug("Creating postal code: {}".format(item))
                pc = PostalCode(
                    country=country,
                    code=code,
                    name=item['placeName'],
                    region_name=item['admin1Name'],
                    subregion_name=item['admin2Name'],
                    district_name=item['admin3Name'])

            if pc.region_name != '':
                try:
                    with transaction.atomic():
                        pc.region = Region.objects.get(
                            Q(name_std__iexact=pc.region_name) |
                            Q(name__iexact=pc.region_name),
                            country=pc.country)
                except Region.DoesNotExist:
                    pc.region = None
            else:
                pc.region = None

            if pc.subregion_name != '':
                try:
                    with transaction.atomic():
                        pc.subregion = Subregion.objects.get(
                            Q(region__name_std__iexact=pc.region_name) |
                            Q(region__name__iexact=pc.region_name),
                            Q(name_std__iexact=pc.subregion_name) |
                            Q(name__iexact=pc.subregion_name),
                            region__country=pc.country)
                except Subregion.DoesNotExist:
                    pc.subregion = None
            else:
                pc.subregion = None

            if pc.district_name != '':
                try:
                    with transaction.atomic():
                        pc.district = District.objects.get(
                            Q(city__region__name_std__iexact=pc.region_name) |
                            Q(city__region__name__iexact=pc.region_name),
                            Q(name_std__iexact=pc.district_name) |
                            Q(name__iexact=pc.district_name),
                            city__country=pc.country)
                except District.MultipleObjectsReturned as e:
                    self.logger.debug("item: {}\ndistricts: {}".format(
                        item,
                        District.objects.filter(
                            Q(city__region__name_std__iexact=pc.region_name) |
                            Q(city__region__name__iexact=pc.region_name),
                            Q(name_std__iexact=pc.district_name) |
                            Q(name__iexact=pc.district_name),
                            city__country=pc.country).values_list('id', flat=True)))
                    # If they're both part of the same city
                    if District.objects.filter(Q(city__region__name_std__iexact=pc.region_name) |
                                               Q(city__region__name__iexact=pc.region_name),
                                               Q(name_std__iexact=pc.district_name) |
                                               Q(name__iexact=pc.district_name),
                                               city__country=pc.country)\
                               .values_list('city').distinct().count() == 1:
                        # Use the one with the lower ID
                        pc.district = District.objects.filter(
                            Q(city__region__name_std__iexact=pc.region_name) |
                            Q(city__region__name__iexact=pc.region_name),
                            Q(name_std__iexact=pc.district_name) |
                            Q(name__iexact=pc.district_name),
                            city__country=pc.country).order_by('city__id').first()

                        districts_to_delete.append(District.objects.filter(
                            Q(city__region__name_std__iexact=pc.region_name) |
                            Q(city__region__name__iexact=pc.region_name),
                            Q(name_std__iexact=pc.district_name) |
                            Q(name__iexact=pc.district_name),
                            city__country=pc.country).order_by('city__id').last().id)
                    else:
                        raise e
                except District.DoesNotExist:
                    pc.district = None
            else:
                pc.district = None

            if pc.district is not None:
                pc.city = pc.district.city
            else:
                pc.city = None

            try:
                pc.location = Point(float(item['longitude']), float(item['latitude']))
            except Exception as e:
                self.logger.warning("Postal code %s (%s) - invalid location ('%s', '%s'): %s",
                                    pc.code, pc.country, item['longitude'],
                                    item['latitude'], str(e))
                pc.location = None

            pc.save()

            if not self.call_hook('postal_code_post', pc, item):
                continue

            self.logger.debug("Added postal code: %s, %s", pc.country, pc)

        if num_existing_postal_codes > 0 and max(query_statistics) > 0:
            width = int(math.log10(max(query_statistics)))

            stats_str = ""
            for i, count in enumerate(query_statistics):
                stats_str = "{{}}\n{{:>2}} [{{:>{}}}]: {{}}".format(width)\
                    .format(stats_str, i, count,
                            ''.join(['=' for i in range(count)]))

                self.logger.info("Postal code query statistics:\n{}".format(stats_str))

        if districts_to_delete:
            self.logger.debug('districts to delete:\n{}'.format(districts_to_delete))

    def flush_country(self):
        self.logger.info("Flushing country data")
        Country.objects.all().delete()

    def flush_region(self):
        self.logger.info("Flushing region data")
        Region.objects.all().delete()

    def flush_subregion(self):
        self.logger.info("Flushing subregion data")
        Subregion.objects.all().delete()

    def flush_city(self):
        self.logger.info("Flushing city data")
        City.objects.all().delete()

    def flush_district(self):
        self.logger.info("Flushing district data")
        District.objects.all().delete()

    def flush_postal_code(self):
        self.logger.info("Flushing postal code data")
        PostalCode.objects.all().delete()

    def flush_alt_name(self):
        self.logger.info("Flushing alternate name data")
        for type_ in (Country, Region, Subregion, City, District, PostalCode):
            plural_type_name = type_.__name__ if type_.__name__[-1] != 'y' else '{}ies'.format(type_.__name__[:-1])
            for obj in tqdm(type_.objects.all(), total=type_.objects.count(),
                            desc="Flushing alternative names for {}".format(
                                plural_type_name)):
                obj.alt_names.all().delete()



from django.contrib import admin

import swapper

from .models import (Continent, Country, Region, Subregion, City, District,
                     PostalCode, AlternativeName)


class CitiesAdmin(admin.ModelAdmin):
    raw_id_fields = ['alt_names']


class ContinentAdmin(CitiesAdmin):
    list_display = ['name', 'code']


class CountryAdmin(CitiesAdmin):
    list_display = ['name', 'code', 'code3', 'tld', 'phone', 'continent', 'area', 'population']
    search_fields = ['name', 'code', 'code3', 'tld', 'phone']


class RegionAdmin(CitiesAdmin):
    ordering = ['name_std']
    list_display = ['name_std', 'code', 'country']
    search_fields = ['name', 'name_std', 'code']


class SubregionAdmin(CitiesAdmin):
    ordering = ['name_std']
    list_display = ['name_std', 'code', 'region']
    search_fields = ['name', 'name_std', 'code']
    raw_id_fields = ['alt_names', 'region']


class CityAdmin(CitiesAdmin):
    ordering = ['name_std']
    list_display = ['name_std', 'subregion', 'region', 'country', 'population']
    search_fields = ['name', 'name_std']
    raw_id_fields = ['alt_names', 'region', 'subregion']


class DistrictAdmin(CitiesAdmin):
    raw_id_fields = ['alt_names', 'city']
    list_display = ['name_std', 'city']
    search_fields = ['name', 'name_std']


class AltNameAdmin(admin.ModelAdmin):
    ordering = ['name']
    list_display = ['name', 'language_code', 'is_preferred', 'is_short', 'is_historic']
    list_filter = ['is_preferred', 'is_short', 'is_historic', 'language_code']
    search_fields = ['name']


class PostalCodeAdmin(CitiesAdmin):
    ordering = ['code']
    list_display = ['code', 'subregion_name', 'region_name', 'country']
    search_fields = ['code', 'country__name', 'region_name', 'subregion_name']


if not swapper.is_swapped('cities', 'Continent'):
    admin.site.register(Continent, ContinentAdmin)
if not swapper.is_swapped('cities', 'Country'):
    admin.site.register(Country, CountryAdmin)
admin.site.register(Region, RegionAdmin)
admin.site.register(Subregion, SubregionAdmin)
if not swapper.is_swapped('cities', 'City'):
    admin.site.register(City, CityAdmin)
admin.site.register(District, DistrictAdmin)
admin.site.register(AlternativeName, AltNameAdmin)
admin.site.register(PostalCode, PostalCodeAdmin)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.2 on 2016-03-29 19:49
from __future__ import unicode_literals

import django.contrib.gis.db.models.fields
from django.db import migrations, models
import django.db.models.deletion

import swapper

from ..conf import CONTINENT_DATA


def get_model(apps, name):
    model_tuple = swapper.split(swapper.get_model_name('cities', name))
    return apps.get_model(*model_tuple)


def add_continents(apps, schema_editor):
    Continent = get_model(apps, 'Continent')

    for ccode, cdata in CONTINENT_DATA.items():
        try:
            c = Continent.objects.get(code=ccode)
        except Continent.DoesNotExist:
            c = Continent()
        c.id = cdata[1]
        c.name = cdata[0]
        c.code = ccode
        c.slug = c.name
        c.save()


def rm_continents(apps, schema_editor):
    # The table is going to be nuked anyway, we just don't want RunPython()
    # to throw an exception on backwards migrations
    pass


def add_continent_fks(apps, schema_editor):
    Country = get_model(apps, 'Country')
    Continent = get_model(apps, 'Continent')

    for continent in Continent.objects.all():
        Country.objects.filter(continent_code=continent.code).update(continent=continent)


def rm_continent_fks(apps, schema_editor):
    Country = get_model(apps, 'Country')
    Continent = get_model(apps, 'Continent')

    for continent in Continent.objects.all():
        Country.objects.filter(continent=continent).update(continent_code=continent.code)


class Migration(migrations.Migration):

    dependencies = [
        ('cities', '0001_initial'),
        swapper.dependency('cities', 'Country'),
    ]

    operations = [
        migrations.CreateModel(
            name='Continent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(db_index=True, max_length=200, verbose_name='ascii name')),
                ('slug', models.CharField(max_length=200, unique=True)),
                ('code', models.CharField(db_index=True, max_length=2, unique=True)),
            ],
            options={
                'abstract': False,
                'swappable': swapper.swappable_setting('cities', 'Continent'),
            },
        ),
        migrations.AddField(
            model_name='continent',
            name='alt_names',
            field=models.ManyToManyField(related_name='cities_continents', to='cities.AlternativeName'),
        ),
        migrations.RenameField(
            model_name='country',
            old_name='continent',
            new_name='continent_code',
        ),
        migrations.AddField(
            model_name='country',
            name='continent',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='countries', to=swapper.get_model_name('cities', 'Continent')),
        ),
        migrations.RunPython(add_continents, rm_continents),
        migrations.RunPython(add_continent_fks, rm_continent_fks),
        migrations.RemoveField(
            model_name='country',
            name='continent_code',
        ),
    ]



import re
import six
import sys
import unicodedata
from math import radians, sin, cos, acos
from django import VERSION as DJANGO_VERSION
try:
    from django.utils.encoding import force_unicode as force_text
except (NameError, ImportError):
    from django.utils.encoding import force_text
from django.utils.safestring import mark_safe, SafeText


if sys.version_info < (3, 0):
    unicode_func = unicode  # noqa: F821
else:
    unicode_func = str


# GEO DISTANCE

earth_radius_km = 6371.009


def geo_distance(a, b):
    """Distance between two geo points in km. (p.x = long, p.y = lat)"""
    a_y = radians(a.y)
    b_y = radians(b.y)
    delta_x = radians(a.x - b.x)
    cos_x = (sin(a_y) * sin(b_y) +
             cos(a_y) * cos(b_y) * cos(delta_x))
    return acos(cos_x) * earth_radius_km


# SLUGIFY REGEXES

to_und_rgx = re.compile(r"[']", re.UNICODE)
slugify_rgx = re.compile(r'[^-\w._~]', re.UNICODE)
multi_dash_rgx = re.compile(r'-{2,}', re.UNICODE)
dash_und_rgx = re.compile(r'[-_]_', re.UNICODE)
und_dash_rgx = re.compile(r'[-_]-', re.UNICODE)
starting_chars_rgx = re.compile(r'^[-._]*', re.UNICODE)
ending_chars_rgx = re.compile(r'[-._]*$', re.UNICODE)


def default_slugify(obj, value):
    if value is None:
        return None

    value = force_text(unicode_func(value))
    value = unicodedata.normalize('NFKC', value.strip())
    value = re.sub(to_und_rgx, '_', value)
    value = re.sub(slugify_rgx, '-', value)
    value = re.sub(multi_dash_rgx, '-', value)
    value = re.sub(dash_und_rgx, '_', value)
    value = re.sub(und_dash_rgx, '_', value)
    value = re.sub(starting_chars_rgx, '', value)
    value = re.sub(ending_chars_rgx, '', value)
    return mark_safe(value)


if DJANGO_VERSION < (1, 10):
    from django.utils.functional import allow_lazy
    default_slugify = allow_lazy(default_slugify, six.text_type, SafeText)
else:
    from django.utils.functional import keep_lazy
    default_slugify = keep_lazy(six.text_type, SafeText)(default_slugify)


# DJANGO BACKWARDS-COMPATIBLE PATTERNS

def patterns(prefix, *args):
    if DJANGO_VERSION < (1, 9):
        from django.conf.urls import patterns as django_patterns
        return django_patterns(prefix, *args)
    elif prefix != '':
        raise Exception("You need to update your URLConf to be a list of URL "
                        "objects")
    else:
        return list(args)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.2 on 2016-10-23 23:39
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion

import swapper


class Migration(migrations.Migration):

    dependencies = [
        ('cities', '0004_rename_languages_to_language_codes'),
        swapper.dependency('cities', 'City'),
    ]

    operations = [
        migrations.AddField(
            model_name='postalcode',
            name='city',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='postal_codes', to=swapper.get_model_name('cities', 'City')),
        ),
        migrations.AddField(
            model_name='postalcode',
            name='district',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='postal_codes', to='cities.District'),
        ),
        migrations.AddField(
            model_name='postalcode',
            name='region',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='postal_codes', to='cities.Region'),
        ),
        migrations.AddField(
            model_name='postalcode',
            name='subregion',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='postal_codes', to='cities.Subregion'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.2 on 2016-10-23 23:24
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion

import swapper


class Migration(migrations.Migration):

    dependencies = [
        ('cities', '0002_continent_models_and_foreign_keys'),
        swapper.dependency('cities', 'Continent'),
        swapper.dependency('cities', 'Country'),
        swapper.dependency('cities', 'City'),
    ]

    operations = [
        migrations.AlterField(
            model_name='city',
            name='country',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='cities', to=swapper.get_model_name('cities', 'Country')),
        ),
        migrations.AlterField(
            model_name='city',
            name='region',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='cities', to='cities.Region'),
        ),
        migrations.AlterField(
            model_name='city',
            name='subregion',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='cities', to='cities.Subregion'),
        ),
        migrations.AlterField(
            model_name='continent',
            name='alt_names',
            field=models.ManyToManyField(to='cities.AlternativeName'),
        ),
        migrations.AlterField(
            model_name='continent',
            name='slug',
            field=models.CharField(max_length=200),
        ),
        migrations.AlterField(
            model_name='country',
            name='tld',
            field=models.CharField(max_length=5, verbose_name='TLD'),
        ),
        migrations.AlterField(
            model_name='district',
            name='city',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='districts', to=swapper.get_model_name('cities', 'City')),
        ),
        migrations.AlterField(
            model_name='region',
            name='country',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='regions', to=swapper.get_model_name('cities', 'Country')),
        ),
        migrations.AlterField(
            model_name='subregion',
            name='region',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='subregions', to='cities.Region'),
        ),
    ]



# -*- coding: utf-8 -*-
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'Country'
        db.create_table(u'cities_country', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=200, db_index=True)),
            ('slug', self.gf('django.db.models.fields.CharField')(max_length=200)),
            ('code', self.gf('django.db.models.fields.CharField')(max_length=2, db_index=True)),
            ('code3', self.gf('django.db.models.fields.CharField')(max_length=3, db_index=True)),
            ('population', self.gf('django.db.models.fields.IntegerField')()),
            ('area', self.gf('django.db.models.fields.IntegerField')(null=True)),
            ('currency', self.gf('django.db.models.fields.CharField')(max_length=3, null=True)),
            ('currency_name', self.gf('django.db.models.fields.CharField')(max_length=50, null=True)),
            ('languages', self.gf('django.db.models.fields.CharField')(max_length=250, null=True)),
            ('phone', self.gf('django.db.models.fields.CharField')(max_length=20)),
            ('continent', self.gf('django.db.models.fields.CharField')(max_length=2)),
            ('tld', self.gf('django.db.models.fields.CharField')(max_length=5)),
            ('capital', self.gf('django.db.models.fields.CharField')(max_length=100)),
        ))
        db.send_create_signal(u'cities', ['Country'])

        # Adding M2M table for field alt_names on 'Country'
        m2m_table_name = db.shorten_name(u'cities_country_alt_names')
        db.create_table(m2m_table_name, (
            ('id', models.AutoField(verbose_name='ID', primary_key=True, auto_created=True)),
            ('country', models.ForeignKey(orm[u'cities.country'], null=False)),
            ('alternativename', models.ForeignKey(orm[u'cities.alternativename'], null=False))
        ))
        db.create_unique(m2m_table_name, ['country_id', 'alternativename_id'])

        # Adding M2M table for field neighbours on 'Country'
        m2m_table_name = db.shorten_name(u'cities_country_neighbours')
        db.create_table(m2m_table_name, (
            ('id', models.AutoField(verbose_name='ID', primary_key=True, auto_created=True)),
            ('from_country', models.ForeignKey(orm[u'cities.country'], null=False)),
            ('to_country', models.ForeignKey(orm[u'cities.country'], null=False))
        ))
        db.create_unique(m2m_table_name, ['from_country_id', 'to_country_id'])

        # Adding model 'Region'
        db.create_table(u'cities_region', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=200, db_index=True)),
            ('slug', self.gf('django.db.models.fields.CharField')(max_length=200)),
            ('name_std', self.gf('django.db.models.fields.CharField')(max_length=200, db_index=True)),
            ('code', self.gf('django.db.models.fields.CharField')(max_length=200, db_index=True)),
            ('country', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['cities.Country'])),
        ))
        db.send_create_signal(u'cities', ['Region'])

        # Adding M2M table for field alt_names on 'Region'
        m2m_table_name = db.shorten_name(u'cities_region_alt_names')
        db.create_table(m2m_table_name, (
            ('id', models.AutoField(verbose_name='ID', primary_key=True, auto_created=True)),
            ('region', models.ForeignKey(orm[u'cities.region'], null=False)),
            ('alternativename', models.ForeignKey(orm[u'cities.alternativename'], null=False))
        ))
        db.create_unique(m2m_table_name, ['region_id', 'alternativename_id'])

        # Adding model 'Subregion'
        db.create_table(u'cities_subregion', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=200, db_index=True)),
            ('slug', self.gf('django.db.models.fields.CharField')(max_length=200)),
            ('name_std', self.gf('django.db.models.fields.CharField')(max_length=200, db_index=True)),
            ('code', self.gf('django.db.models.fields.CharField')(max_length=200, db_index=True)),
            ('region', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['cities.Region'])),
        ))
        db.send_create_signal(u'cities', ['Subregion'])

        # Adding M2M table for field alt_names on 'Subregion'
        m2m_table_name = db.shorten_name(u'cities_subregion_alt_names')
        db.create_table(m2m_table_name, (
            ('id', models.AutoField(verbose_name='ID', primary_key=True, auto_created=True)),
            ('subregion', models.ForeignKey(orm[u'cities.subregion'], null=False)),
            ('alternativename', models.ForeignKey(orm[u'cities.alternativename'], null=False))
        ))
        db.create_unique(m2m_table_name, ['subregion_id', 'alternativename_id'])

        # Adding model 'City'
        db.create_table(u'cities_city', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=200, db_index=True)),
            ('slug', self.gf('django.db.models.fields.CharField')(max_length=200)),
            ('name_std', self.gf('django.db.models.fields.CharField')(max_length=200, db_index=True)),
            ('location', self.gf('django.contrib.gis.db.models.fields.PointField')()),
            ('population', self.gf('django.db.models.fields.IntegerField')()),
            ('region', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['cities.Region'], null=True, blank=True)),
            ('subregion', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['cities.Subregion'], null=True, blank=True)),
            ('country', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['cities.Country'])),
            ('elevation', self.gf('django.db.models.fields.IntegerField')(null=True)),
            ('kind', self.gf('django.db.models.fields.CharField')(max_length=10)),
            ('timezone', self.gf('django.db.models.fields.CharField')(max_length=40)),
        ))
        db.send_create_signal(u'cities', ['City'])

        # Adding M2M table for field alt_names on 'City'
        m2m_table_name = db.shorten_name(u'cities_city_alt_names')
        db.create_table(m2m_table_name, (
            ('id', models.AutoField(verbose_name='ID', primary_key=True, auto_created=True)),
            ('city', models.ForeignKey(orm[u'cities.city'], null=False)),
            ('alternativename', models.ForeignKey(orm[u'cities.alternativename'], null=False))
        ))
        db.create_unique(m2m_table_name, ['city_id', 'alternativename_id'])

        # Adding model 'District'
        db.create_table(u'cities_district', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=200, db_index=True)),
            ('slug', self.gf('django.db.models.fields.CharField')(max_length=200)),
            ('name_std', self.gf('django.db.models.fields.CharField')(max_length=200, db_index=True)),
            ('location', self.gf('django.contrib.gis.db.models.fields.PointField')()),
            ('population', self.gf('django.db.models.fields.IntegerField')()),
            ('city', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['cities.City'])),
        ))
        db.send_create_signal(u'cities', ['District'])

        # Adding M2M table for field alt_names on 'District'
        m2m_table_name = db.shorten_name(u'cities_district_alt_names')
        db.create_table(m2m_table_name, (
            ('id', models.AutoField(verbose_name='ID', primary_key=True, auto_created=True)),
            ('district', models.ForeignKey(orm[u'cities.district'], null=False)),
            ('alternativename', models.ForeignKey(orm[u'cities.alternativename'], null=False))
        ))
        db.create_unique(m2m_table_name, ['district_id', 'alternativename_id'])

        # Adding model 'AlternativeName'
        db.create_table(u'cities_alternativename', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=256)),
            ('language', self.gf('django.db.models.fields.CharField')(max_length=100)),
            ('is_preferred', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('is_short', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('is_colloquial', self.gf('django.db.models.fields.BooleanField')(default=False)),
        ))
        db.send_create_signal(u'cities', ['AlternativeName'])

        # Adding model 'PostalCode'
        db.create_table(u'cities_postalcode', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=200, db_index=True)),
            ('slug', self.gf('django.db.models.fields.CharField')(max_length=200)),
            ('code', self.gf('django.db.models.fields.CharField')(max_length=20)),
            ('location', self.gf('django.contrib.gis.db.models.fields.PointField')()),
            ('country', self.gf('django.db.models.fields.related.ForeignKey')(related_name='postal_codes', to=orm['cities.Country'])),
            ('region_name', self.gf('django.db.models.fields.CharField')(max_length=100, db_index=True)),
            ('subregion_name', self.gf('django.db.models.fields.CharField')(max_length=100, db_index=True)),
            ('district_name', self.gf('django.db.models.fields.CharField')(max_length=100, db_index=True)),
        ))
        db.send_create_signal(u'cities', ['PostalCode'])

        # Adding M2M table for field alt_names on 'PostalCode'
        m2m_table_name = db.shorten_name(u'cities_postalcode_alt_names')
        db.create_table(m2m_table_name, (
            ('id', models.AutoField(verbose_name='ID', primary_key=True, auto_created=True)),
            ('postalcode', models.ForeignKey(orm[u'cities.postalcode'], null=False)),
            ('alternativename', models.ForeignKey(orm[u'cities.alternativename'], null=False))
        ))
        db.create_unique(m2m_table_name, ['postalcode_id', 'alternativename_id'])

    def backwards(self, orm):
        # Deleting model 'Country'
        db.delete_table(u'cities_country')

        # Removing M2M table for field alt_names on 'Country'
        db.delete_table(db.shorten_name(u'cities_country_alt_names'))

        # Removing M2M table for field neighbours on 'Country'
        db.delete_table(db.shorten_name(u'cities_country_neighbours'))

        # Deleting model 'Region'
        db.delete_table(u'cities_region')

        # Removing M2M table for field alt_names on 'Region'
        db.delete_table(db.shorten_name(u'cities_region_alt_names'))

        # Deleting model 'Subregion'
        db.delete_table(u'cities_subregion')

        # Removing M2M table for field alt_names on 'Subregion'
        db.delete_table(db.shorten_name(u'cities_subregion_alt_names'))

        # Deleting model 'City'
        db.delete_table(u'cities_city')

        # Removing M2M table for field alt_names on 'City'
        db.delete_table(db.shorten_name(u'cities_city_alt_names'))

        # Deleting model 'District'
        db.delete_table(u'cities_district')

        # Removing M2M table for field alt_names on 'District'
        db.delete_table(db.shorten_name(u'cities_district_alt_names'))

        # Deleting model 'AlternativeName'
        db.delete_table(u'cities_alternativename')

        # Deleting model 'PostalCode'
        db.delete_table(u'cities_postalcode')

        # Removing M2M table for field alt_names on 'PostalCode'
        db.delete_table(db.shorten_name(u'cities_postalcode_alt_names'))

    models = {
        u'cities.alternativename': {
            'Meta': {'object_name': 'AlternativeName'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_colloquial': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'is_preferred': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'is_short': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'language': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '256'})
        },
        u'cities.city': {
            'Meta': {'object_name': 'City'},
            'alt_names': ('django.db.models.fields.related.ManyToManyField', [], {'to': u"orm['cities.AlternativeName']", 'symmetrical': 'False'}),
            'country': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['cities.Country']"}),
            'elevation': ('django.db.models.fields.IntegerField', [], {'null': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'kind': ('django.db.models.fields.CharField', [], {'max_length': '10'}),
            'location': ('django.contrib.gis.db.models.fields.PointField', [], {}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '200', 'db_index': 'True'}),
            'name_std': ('django.db.models.fields.CharField', [], {'max_length': '200', 'db_index': 'True'}),
            'population': ('django.db.models.fields.IntegerField', [], {}),
            'region': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['cities.Region']", 'null': 'True', 'blank': 'True'}),
            'slug': ('django.db.models.fields.CharField', [], {'max_length': '200'}),
            'subregion': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['cities.Subregion']", 'null': 'True', 'blank': 'True'}),
            'timezone': ('django.db.models.fields.CharField', [], {'max_length': '40'})
        },
        u'cities.country': {
            'Meta': {'ordering': "['name']", 'object_name': 'Country'},
            'alt_names': ('django.db.models.fields.related.ManyToManyField', [], {'to': u"orm['cities.AlternativeName']", 'symmetrical': 'False'}),
            'area': ('django.db.models.fields.IntegerField', [], {'null': 'True'}),
            'capital': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'code': ('django.db.models.fields.CharField', [], {'max_length': '2', 'db_index': 'True'}),
            'code3': ('django.db.models.fields.CharField', [], {'max_length': '3', 'db_index': 'True'}),
            'continent': ('django.db.models.fields.CharField', [], {'max_length': '2'}),
            'currency': ('django.db.models.fields.CharField', [], {'max_length': '3', 'null': 'True'}),
            'currency_name': ('django.db.models.fields.CharField', [], {'max_length': '50', 'null': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'languages': ('django.db.models.fields.CharField', [], {'max_length': '250', 'null': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '200', 'db_index': 'True'}),
            'neighbours': ('django.db.models.fields.related.ManyToManyField', [], {'related_name': "'neighbours_rel_+'", 'to': u"orm['cities.Country']"}),
            'phone': ('django.db.models.fields.CharField', [], {'max_length': '20'}),
            'population': ('django.db.models.fields.IntegerField', [], {}),
            'slug': ('django.db.models.fields.CharField', [], {'max_length': '200'}),
            'tld': ('django.db.models.fields.CharField', [], {'max_length': '5'})
        },
        u'cities.district': {
            'Meta': {'object_name': 'District'},
            'alt_names': ('django.db.models.fields.related.ManyToManyField', [], {'to': u"orm['cities.AlternativeName']", 'symmetrical': 'False'}),
            'city': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['cities.City']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'location': ('django.contrib.gis.db.models.fields.PointField', [], {}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '200', 'db_index': 'True'}),
            'name_std': ('django.db.models.fields.CharField', [], {'max_length': '200', 'db_index': 'True'}),
            'population': ('django.db.models.fields.IntegerField', [], {}),
            'slug': ('django.db.models.fields.CharField', [], {'max_length': '200'})
        },
        u'cities.postalcode': {
            'Meta': {'object_name': 'PostalCode'},
            'alt_names': ('django.db.models.fields.related.ManyToManyField', [], {'to': u"orm['cities.AlternativeName']", 'symmetrical': 'False'}),
            'code': ('django.db.models.fields.CharField', [], {'max_length': '20'}),
            'country': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'postal_codes'", 'to': u"orm['cities.Country']"}),
            'district_name': ('django.db.models.fields.CharField', [], {'max_length': '100', 'db_index': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'location': ('django.contrib.gis.db.models.fields.PointField', [], {}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '200', 'db_index': 'True'}),
            'region_name': ('django.db.models.fields.CharField', [], {'max_length': '100', 'db_index': 'True'}),
            'slug': ('django.db.models.fields.CharField', [], {'max_length': '200'}),
            'subregion_name': ('django.db.models.fields.CharField', [], {'max_length': '100', 'db_index': 'True'})
        },
        u'cities.region': {
            'Meta': {'object_name': 'Region'},
            'alt_names': ('django.db.models.fields.related.ManyToManyField', [], {'to': u"orm['cities.AlternativeName']", 'symmetrical': 'False'}),
            'code': ('django.db.models.fields.CharField', [], {'max_length': '200', 'db_index': 'True'}),
            'country': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['cities.Country']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '200', 'db_index': 'True'}),
            'name_std': ('django.db.models.fields.CharField', [], {'max_length': '200', 'db_index': 'True'}),
            'slug': ('django.db.models.fields.CharField', [], {'max_length': '200'})
        },
        u'cities.subregion': {
            'Meta': {'object_name': 'Subregion'},
            'alt_names': ('django.db.models.fields.related.ManyToManyField', [], {'to': u"orm['cities.AlternativeName']", 'symmetrical': 'False'}),
            'code': ('django.db.models.fields.CharField', [], {'max_length': '200', 'db_index': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '200', 'db_index': 'True'}),
            'name_std': ('django.db.models.fields.CharField', [], {'max_length': '200', 'db_index': 'True'}),
            'region': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['cities.Region']"}),
            'slug': ('django.db.models.fields.CharField', [], {'max_length': '200'})
        }
    }

    complete_apps = ['cities']



# -*- coding: utf-8 -*-
# Generated by Django 1.10.2 on 2016-12-20 08:10
from __future__ import unicode_literals

from django.db import migrations, models


UNIQUE_SLUG_MODELS = ['Continent', 'Country', 'Region', 'Subregion', 'District',
                      'City', 'PostalCode', 'AlternativeName']


class Migration(migrations.Migration):

    dependencies = [
        ('cities', '0009_add_slug_fields_to_models'),
    ]

    operations = [
        migrations.AlterField(
            model_name='country',
            name='code',
            field=models.CharField(db_index=True, max_length=2, unique=True),
        ),
        migrations.AlterField(
            model_name='country',
            name='code3',
            field=models.CharField(db_index=True, max_length=3, unique=True),
        ),
        migrations.AlterUniqueTogether(
            name='city',
            unique_together=set([('country', 'region', 'subregion', 'id', 'name')]),
        ),
        migrations.AlterUniqueTogether(
            name='district',
            unique_together=set([('city', 'name')]),
        ),
        migrations.AlterUniqueTogether(
            name='postalcode',
            unique_together=set([
                ('country', 'region_name', 'subregion_name', 'district_name', 'name', 'id', 'code'),
                ('country', 'region', 'subregion', 'city', 'district', 'name', 'id', 'code'),
            ]),
        ),
        migrations.AlterUniqueTogether(
            name='region',
            unique_together=set([('country', 'name')]),
        ),
        migrations.AlterUniqueTogether(
            name='subregion',
            unique_together=set([('region', 'id', 'name')]),
        ),
    ]

    # The previous migration created the slug field as
    #
    # CharField(max_length=255, unique=True)
    #
    # but I removed the unique=True for new users. This checks the previous
    # state of the slug field on the models, and if the field has unique=True
    # (eg: existing users), it removes the unique attribute. For slug fields
    # that don't have unique=True (eg: new users), it skips them.
    def database_forwards(self, app_label, schema_editor, from_state, to_state):
        # Before we do any database operations, add the appropriate slug field
        # modifications to the migration operations.
        self.operations + [
            migrations.AlterField(
                model_name=model_name.lower(),
                name='slug',
                field=models.CharField(max_length=255))
            for model_name in UNIQUE_SLUG_MODELS
            # This is the check that matters. It checks the result of the
            # previous 0009 migration (which may or may not have unique=True
            # set on slug fields) for the unique attribute, and skips it if
            # they don't.
            if from_state.apps.get_model(app_label, self.model_name)._meta.get_field('slug').unique
        ]

        super(Migration, self).database_forwards(app_label, schema_editor, from_state, to_state)



from setuptools import setup, find_packages
import codecs
import os


def read(fname):
    """
    Utility function to read the README file.
    Used for the long_description. It's nice, because now (1) we have a top level
    README file and (2) it's easier to type in the README file than to put a raw
    string in below ...
    """
    return codecs.open(os.path.join(os.path.dirname(__file__), fname), encoding='utf-8').read()


setup(
    name='django-cities',
    version='0.5.0.3',
    description='Place models and worldwide place data for Django',
    author='Ben Dowling',
    author_email='ben.m.dowling@gmail.com',
    url='https://github.com/coderholic/django-cities',
    packages=find_packages(exclude=['example']),
    install_requires=[
        'django-model-utils',
        'six',
        'swapper',
        'tqdm',
    ],
    include_package_data=True,
    zip_safe=False,
    long_description=read('README.md'),
    tests_require=[
        'flake8',
        'psycopg2',
        'tox',
    ],
    license="MIT",
    keywords="django cities countries regions postal codes geonames",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



from random import choice
from string import ascii_uppercase, digits

try:
    from django.utils.encoding import force_unicode as force_text
except (NameError, ImportError):
    from django.utils.encoding import force_text

from django.db import transaction
from django.utils.encoding import python_2_unicode_compatible
from django.contrib.gis.db import models
from django.contrib.gis.geos import Point

from model_utils import Choices
import swapper

from .conf import (ALTERNATIVE_NAME_TYPES, SLUGIFY_FUNCTION)
from .managers import AlternativeNameManager
from .util import unicode_func

__all__ = [
    'Point', 'Continent', 'Country', 'Region', 'Subregion', 'City', 'District',
    'PostalCode', 'AlternativeName',
]


slugify_func = SLUGIFY_FUNCTION


class SlugModel(models.Model):
    slug = models.CharField(blank=True, max_length=255, null=True)

    class Meta:
        abstract = True

    def slugify(self):
        raise NotImplementedError("Subclasses of Place must implement slugify()")

    def save(self, *args, **kwargs):
        self.slug = slugify_func(self, self.slugify())
        # If the slug contains the object's ID and we are creating a new object,
        # save it twice: once to get an ID, another to set the object's slug
        if self.slug is None and getattr(self, 'slug_contains_id', False):
            with transaction.atomic():
                # We first give a randomized slug with a prefix just in case
                # users need to find invalid slugs
                self.slug = 'invalid-{}'.format(''.join(choice(ascii_uppercase + digits) for i in range(20)))
                super(SlugModel, self).save(*args, **kwargs)
                self.slug = slugify_func(self, self.slugify())
                super(SlugModel, self).save(*args, **kwargs)
        else:
            # This is a performance optimization - we avoid the transaction if
            # the self.slug is not None
            super(SlugModel, self).save(*args, **kwargs)


@python_2_unicode_compatible
class Place(models.Model):
    name = models.CharField(max_length=200, db_index=True, verbose_name="ascii name")
    alt_names = models.ManyToManyField('AlternativeName')

    objects = models.GeoManager()

    class Meta:
        abstract = True

    @property
    def hierarchy(self):
        """Get hierarchy, root first"""
        lst = self.parent.hierarchy if self.parent else []
        lst.append(self)
        return lst

    def get_absolute_url(self):
        return "/".join([place.slug for place in self.hierarchy])

    def __str__(self):
        return force_text(self.name)

    def save(self, *args, **kwargs):
        if hasattr(self, 'clean'):
            self.clean()
        super(Place, self).save(*args, **kwargs)


class BaseContinent(Place, SlugModel):
    code = models.CharField(max_length=2, unique=True, db_index=True)

    def __str__(self):
        return force_text(self.name)

    class Meta:
        abstract = True

    def slugify(self):
        return self.name


class Continent(BaseContinent):
    class Meta(BaseContinent.Meta):
        swappable = swapper.swappable_setting('cities', 'Continent')


class BaseCountry(Place, SlugModel):
    code = models.CharField(max_length=2, db_index=True, unique=True)
    code3 = models.CharField(max_length=3, db_index=True, unique=True)
    population = models.IntegerField()
    area = models.IntegerField(null=True)
    currency = models.CharField(max_length=3, null=True)
    currency_name = models.CharField(max_length=50, blank=True, null=True)
    currency_symbol = models.CharField(max_length=31, blank=True, null=True)
    language_codes = models.CharField(max_length=250, null=True)
    phone = models.CharField(max_length=20)
    continent = models.ForeignKey(swapper.get_model_name('cities', 'Continent'),
                                  null=True, related_name='countries')
    tld = models.CharField(max_length=5, verbose_name='TLD')
    postal_code_format = models.CharField(max_length=127)
    postal_code_regex = models.CharField(max_length=255)
    capital = models.CharField(max_length=100)
    neighbours = models.ManyToManyField("self")

    class Meta:
        abstract = True
        ordering = ['name']
        verbose_name_plural = "countries"

    @property
    def parent(self):
        return None

    def __str__(self):
        return force_text(self.name)

    def clean(self):
        self.tld = self.tld.lower()

    def slugify(self):
        return self.name


class Country(BaseCountry):
    class Meta(BaseCountry.Meta):
        swappable = swapper.swappable_setting('cities', 'Country')


class Region(Place, SlugModel):
    name_std = models.CharField(max_length=200, db_index=True, verbose_name="standard name")
    code = models.CharField(max_length=200, db_index=True)
    country = models.ForeignKey(swapper.get_model_name('cities', 'Country'),
                                related_name='regions')

    class Meta:
        unique_together = (('country', 'name'),)

    @property
    def parent(self):
        return self.country

    def full_code(self):
        return unicode_func(".".join([self.parent.code, self.code]))

    def slugify(self):
        return '{}_({})'.format(
            unicode_func(self.name),
            unicode_func(self.full_code()))


class Subregion(Place, SlugModel):
    slug_contains_id = True

    name_std = models.CharField(max_length=200, db_index=True, verbose_name="standard name")
    code = models.CharField(max_length=200, db_index=True)
    region = models.ForeignKey(Region, related_name='subregions')

    class Meta:
        unique_together = (('region', 'id', 'name'),)

    @property
    def parent(self):
        return self.region

    def full_code(self):
        return ".".join([self.parent.parent.code, self.parent.code, self.code])

    def slugify(self):
        return unicode_func('{}-{}').format(
            unicode_func(self.id),
            unicode_func(self.name))


class BaseCity(Place, SlugModel):
    slug_contains_id = True

    name_std = models.CharField(max_length=200, db_index=True, verbose_name="standard name")
    country = models.ForeignKey(swapper.get_model_name('cities', 'Country'),
                                related_name='cities')
    region = models.ForeignKey(Region, null=True, blank=True, related_name='cities')
    subregion = models.ForeignKey(Subregion, null=True, blank=True, related_name='cities')
    location = models.PointField()
    population = models.IntegerField()
    elevation = models.IntegerField(null=True)
    kind = models.CharField(max_length=10)  # http://www.geonames.org/export/codes.html
    timezone = models.CharField(max_length=40)

    class Meta:
        abstract = True
        unique_together = (('country', 'region', 'subregion', 'id', 'name'),)
        verbose_name_plural = "cities"

    @property
    def parent(self):
        return self.region

    def slugify(self):
        if self.id:
            return '{}-{}'.format(self.id, unicode_func(self.name))
        return None


class City(BaseCity):
    class Meta(BaseCity.Meta):
        swappable = swapper.swappable_setting('cities', 'City')


class District(Place, SlugModel):
    slug_contains_id = True

    name_std = models.CharField(max_length=200, db_index=True, verbose_name="standard name")
    code = models.CharField(blank=True, db_index=True, max_length=200, null=True)
    location = models.PointField()
    population = models.IntegerField()
    city = models.ForeignKey(swapper.get_model_name('cities', 'City'), related_name='districts')

    class Meta:
        unique_together = (('city', 'name'),)

    @property
    def parent(self):
        return self.city

    def slugify(self):
        if self.id:
            return '{}-{}'.format(self.id, unicode_func(self.name))
        return None


@python_2_unicode_compatible
class AlternativeName(SlugModel):
    slug_contains_id = True

    KIND = Choices(*ALTERNATIVE_NAME_TYPES)

    name = models.CharField(max_length=255)
    kind = models.CharField(max_length=4, choices=KIND, default=KIND.name)
    language_code = models.CharField(max_length=100)
    is_preferred = models.BooleanField(default=False)
    is_short = models.BooleanField(default=False)
    is_colloquial = models.BooleanField(default=False)
    is_historic = models.BooleanField(default=False)

    objects = AlternativeNameManager()

    def __str__(self):
        return "%s (%s)" % (force_text(self.name), force_text(self.language_code))

    def slugify(self):
        if self.id:
            return '{}-{}'.format(self.id, unicode_func(self.name))
        return None


@python_2_unicode_compatible
class PostalCode(Place, SlugModel):
    slug_contains_id = True

    code = models.CharField(max_length=20)
    location = models.PointField()

    country = models.ForeignKey(swapper.get_model_name('cities', 'Country'),
                                related_name='postal_codes')

    # Region names for each admin level, region may not exist in DB
    region_name = models.CharField(max_length=100, db_index=True)
    subregion_name = models.CharField(max_length=100, db_index=True)
    district_name = models.CharField(max_length=100, db_index=True)

    region = models.ForeignKey(Region, blank=True, null=True, related_name='postal_codes')
    subregion = models.ForeignKey(Subregion, blank=True, null=True, related_name='postal_codes')
    city = models.ForeignKey(swapper.get_model_name('cities', 'City'),
                             blank=True, null=True, related_name='postal_codes')
    district = models.ForeignKey(District, blank=True, null=True, related_name='postal_codes')

    objects = models.GeoManager()

    class Meta:
        unique_together = (
            ('country', 'region', 'subregion', 'city', 'district', 'name', 'id', 'code'),
            ('country', 'region_name', 'subregion_name', 'district_name', 'name', 'id', 'code'),
        )

    @property
    def parent(self):
        return self.country

    @property
    def name_full(self):
        """Get full name including hierarchy"""
        return force_text(', '.join(reversed(self.names)))

    @property
    def names(self):
        """Get a hierarchy of non-null names, root first"""
        return [e for e in [
            force_text(self.country),
            force_text(self.region_name),
            force_text(self.subregion_name),
            force_text(self.district_name),
            force_text(self.name),
        ] if e]

    def __str__(self):
        return force_text(self.code)

    def slugify(self):
        if self.id:
            return '{}-{}'.format(self.id, unicode_func(self.code))
        return None
